"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const http_1 = require("http");
const spdy_1 = require("spdy");
const app_1 = require("./app");
const config_1 = require("./config");
const database_1 = require("./database");
const logging_1 = require("./logging");
const redis_session_1 = require("./redis-session");
const logger = new logging_1.Logger('server');
function getStructure(config) {
    return __awaiter(this, void 0, void 0, function* () {
        const structure = { server: null, database: null, Session: null, app: null };
        logger.info(`Using NODE_ENV=${config.node_env}`);
        // database
        if (config.database_required) {
            structure.database = yield database_1.getDatabase(config);
        }
        // session
        if (config.session_required) {
            structure.Session = yield redis_session_1.getSession(config);
        }
        // app (express)
        structure.app = yield app_1.getApp(config, structure.Session);
        // server
        structure.server = yield getServer(config, structure.app);
        // reference to the config object as a helper
        structure.config = config;
        return structure;
    });
}
exports.getStructure = getStructure;
function destroyStructure(structure) {
    return __awaiter(this, void 0, void 0, function* () {
        // close the database
        if (structure.database) {
            yield structure.database.destroy();
        }
        // close the session
        if (structure.Session && structure.Session.connection) {
            switch (structure.Session.type) {
                case 'redis':
                    yield structure.Session.connection.quit();
                    break;
                default:
                    logger.log('session type unknown');
            }
        }
        structure = null;
    });
}
exports.destroyStructure = destroyStructure;
function getServer(config, app) {
    return __awaiter(this, void 0, void 0, function* () {
        let server;
        if (config.http2_required) {
            if (!fs_1.existsSync(config.http2_key)) {
                throw new Error(`The key for the https server can't be found`);
            }
            if (!fs_1.existsSync(config.http2_cert)) {
                throw new Error(`The certificate for the https server can't be found`);
            }
            const options = {
                key: fs_1.readFileSync(config.http2_key),
                cert: fs_1.readFileSync(config.http2_cert)
            };
            logger.log('Creating http2 server...');
            server = spdy_1.createServer(options, app);
        }
        else {
            logger.log('Creating basic http server...');
            server = http_1.createServer(app);
        }
        return server;
    });
}
exports.getServer = getServer;
/**
 * @param {string} startupFilepath entrypoint configuration script path.
 *    This script lets you define all the dynamic configuration of your
 *    application, e.g. the routers, the middlewares, the path to a static
 *    configuration file, ...
 */
function startServer(startupFilepath) {
    return __awaiter(this, void 0, void 0, function* () {
        logger.log('Initialising server...');
        try {
            // get defaults & from-configuration-file configurations
            let config = yield config_1.getConfig(startupFilepath);
            // get the structure of all the application/server
            // only one dependency => the configuration object
            const structure = yield getStructure(config);
            structure.server.listen(config.port, (error) => {
                if (error) {
                    throw new Error(error.message);
                }
                logger.success(`Listening http${config.http2_required ? 's' : ''}://${config.local_hostname}:${config.port}`);
            });
            // structure.server.on('error', async (e) => {
            //   logger.error(e.message);
            //   process.exit(1);
            // });
        }
        catch (e) {
            if (!['config', 'database'].includes(e.name)) {
                logger.error(e.message);
            }
            console.error('Something went wrong. exiting.');
            process.exit(1);
        }
    });
}
exports.startServer = startServer;
//# sourceMappingURL=server.js.map